/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include "caches.hpp"

using namespace std;

Cache::Cache() { Mode = Standby; NumberOfFilesOpen = 0; LastVisitedHash = 0xffffffff; MaxNumberOfFilesOpen = 950; };
Cache::~Cache() { };

unsigned int pow2(unsigned int bytes)
{
    unsigned int result = 1, i1;
    for (i1 = 0; i1 < bytes * 8; i1++)
        result = result * 2;

    return result;
}

/* 99.9% copy of SuperFastHash from http://www.azillionmonkeys.com/qed/hash.html */

// START
#include "stdint.h"
#undef get16bits
#if (defined(__GNUC__) && defined(__i386__)) || defined(__WATCOMC__) \
  || defined(_MSC_VER) || defined (__BORLANDC__) || defined (__TURBOC__)
#define get16bits(d) (*((const uint16_t *) (d)))
#endif

#if !defined (get16bits)
#define get16bits(d) ((((uint32_t)(((const uint8_t *)(d))[1])) << 8)\
                       +(uint32_t)(((const uint8_t *)(d))[0]) )
#endif

uint32_t SuperFastHash (const char * data, int len) {
uint32_t hash = len, tmp;
int rem;

    if (len <= 0 || data == NULL) return 0;

    rem = len & 3;
    len >>= 2;

    /* Main loop */
    for (;len > 0; len--) {
        hash  += get16bits (data);
        tmp    = (get16bits (data+2) << 11) ^ hash;
        hash   = (hash << 16) ^ tmp;
        data  += 2*sizeof (uint16_t);
        hash  += hash >> 11;
    }

    /* Handle end cases */
    switch (rem) {
        case 3: hash += get16bits (data);
                hash ^= hash << 16;
                hash ^= data[sizeof (uint16_t)] << 18;
                hash += hash >> 11;
                break;
        case 2: hash += get16bits (data);
                hash ^= hash << 11;
                hash += hash >> 17;
                break;
        case 1: hash += *data;
                hash ^= hash << 10;
                hash += hash >> 1;
    }

    /* Force "avalanching" of final 127 bits */
    hash ^= hash << 3;
    hash += hash >> 5;
    hash ^= hash << 4;
    hash += hash >> 17;
    hash ^= hash << 25;
    hash += hash >> 6;

    return hash;
}
// END

void Cache::hashString(const string &s, unsigned int &result)
{
    result = SuperFastHash(s.c_str(), s.size());
}

bool Cache::newCache(const string &cache_directory, int number_of_files)
{
    if (Mode != Standby)
        return false;

    /* To prevent users from shooting at their leg with stupid
       cache names. */
    string::const_iterator i1;
    for (i1 = cache_directory.begin(); i1 != cache_directory.end(); i1++)
        if (!( ( (*i1) >= 'a' && (*i1) <= 'z' ) ||
              ( (*i1) >= 'A' && (*i1) <= 'Z' ) ||
              ( (*i1) >= '0' && (*i1) <= '9' )))
            return false; /* Invalid character in name */
    
    /* Remove the cache directory (the lazy unportable way that also deletes the files) */
    char dir[1000];
    snprintf(dir, 1000, "rm -rf %s", cache_directory.c_str());
    system(dir);
    
    CacheDirectory = cache_directory;
    
    /* Create new cache */
    mkdir(cache_directory.c_str(), S_IRUSR | S_IWUSR | S_IEXEC);
    
    /* Open cache files for writing */
    CacheStreams.resize(number_of_files, (fstream*) 0);
    
    Mode = Writing;
    return true;
}

void Cache::addSeed(unsigned int hash, unsigned int seed)
{
    fstream* &stream = CacheStreams[hash];
    if (!stream)
    {
        // Close a file if there are too many of them open
        if (MaxNumberOfFilesOpen <= NumberOfFilesOpen)
        {
            if (LastVisitedHash != 0xffffffff)
            {
                CacheStreams[LastVisitedHash]->close();
                delete CacheStreams[LastVisitedHash];
                CacheStreams[LastVisitedHash] = (fstream*) 0;
            }
            else
            {
                unsigned int i1, len = CacheStreams.size();
                for (i1 = 0; i1 < len; i1++)
                    if (CacheStreams[i1])
                    {
                        CacheStreams[i1]->close();
                        delete CacheStreams[i1];
                        CacheStreams[i1] = (fstream*) 0;
                        NumberOfFilesOpen--;
                        break;
                    }
            }
            
        }
        
        char dir[1000];
        snprintf(dir, 1000, "%s/%d", CacheDirectory.c_str(), hash);
        stream = new fstream(dir, fstream::out | fstream::in | fstream::app);
        if (stream->fail())
        {
            cerr << "Failed to open " << dir << endl;
            abort();
        }
        NumberOfFilesOpen++;
    }
    
    stream->write((char*) &seed, sizeof(unsigned int));
    LastVisitedHash = hash;
}

void Cache::closeCache()
{
    unsigned int i1;
    for (i1 = 0; i1 < CacheStreams.size(); i1++)
        if (CacheStreams[i1])
        {
            CacheStreams[i1]->close();
            delete CacheStreams[i1];
            CacheStreams[i1] = (fstream*) 0;
        }
    CacheStreams.clear();
    NumberOfFilesOpen = 0;
    Mode = Standby;
}

bool Cache::openCache(const string &cache_name, int number_of_files)
{
    if (Mode != Standby)
        return false;

    CacheStreams.resize(number_of_files, (fstream*) 0);
    CacheDirectory = cache_name;
    
    Mode = Reading;
    return true;
}

void Cache::readCacheFile(unsigned int value, vector<unsigned int> &values)
{
    if (Mode != Reading)
        return;
    
    /* Calculate, how many values there are with this hash. */
    char dir[1000];
    snprintf(dir, 1000, "%s/%d", CacheDirectory.c_str(), value);
    
    fstream stream(dir, fstream::in);
    if (stream.fail())
    {
        cerr << "Failed to open cache file " << dir << endl;
        abort();
    }
    
    stream.seekg(0, ios::end);
    unsigned long len = stream.tellg() / sizeof(unsigned int);
    stream.seekg(0, ios::beg);
    values.resize(len);
    
    unsigned int seed;
    
    /* read the seeds out of the cache file. */
    unsigned int i1;
    for (i1 = 0; i1 < len; i1++)
    {
        stream.read((char*) &seed, sizeof(seed));
        values[i1] = seed;
    }
    
    stream.close();
}

