/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifndef CACHE_HELPER_H
#define CACHE_HELPER_H

#include <vector>
#include <string>
#include <fstream>

using namespace std;

/*
  Note: The limit on the number of files open
        at the same time is usually 1024.
        Therefore, you can only use 1 byte hashes
        to avoid opening too many files (the current
        implementation opens all the cache files).
        
        This small utility can only work on UNIX-like
        environment.
*/

enum CacheMode { Standby, Reading, Writing };

class Cache
{
    private:
        // No copies
        Cache(const Cache &c) { };
        Cache& operator=(const Cache &c) { return (*this); };
        
        /* All the streams to cache files are here. */
        vector<fstream*> CacheStreams;
        /* Number of files open */
        int NumberOfFilesOpen;
        /* The maximum allowed number of files open */
        int MaxNumberOfFilesOpen;
        /* The last opened cache file*/
        unsigned int LastVisitedHash; 
        
        /* Current mode */
        CacheMode Mode;
        
        string CacheDirectory;
        
    public:
        Cache();
        ~Cache();
        
        /* Hashes string to a shorter value. */
        static void hashString(const string &s, unsigned int &result);
        
        /* Creates a new cache directory (clears it from all files if it existed).
           Returns false if something goes wrong. After successful
           return, cache is in write mode and cannot be read from. 
           The cache name may only contain letters and numbers. */
        bool newCache(const string &cache_directory, int number_of_files);
        /* Add a seed to the cache.
           Cache must be in write mode but safe checking is not done. */
        void addSeed(unsigned int hash, unsigned int seed);
        
        /* When you are done creating or reading the cache, call this.
           Called automatically by destructor. */
        void closeCache();
        
        /* Opens cache for reading. Returns false if something goes wrong. */
        bool openCache(const string &cache_name, int number_of_files);
        /* Reads all values from a cache file that is the same as the hash
           of your string. Note that this reads all the contents to the memory. */
        void readCacheFile(unsigned int hash, vector<unsigned int> &values);
};

#endif

