/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/* This tool needs the seed obtained from other tools.

   1. Find the seed using seed_by_naming or seed_by_discoveries.
   2. Find the current offset in game using seed_by_naming (the end offset)
   3. Use fountain_wish to check how many times you need to #name or walk into
      a wall before quaffing.
   4. Check how many random numbers your wish would consume to see
      if the fountain would dry up (this tool helps you with it).
   5. Do your wish in the game.
   6. Go to step 2 to get more wishes until you are satisfied.
   
   The number of random()s in a wish, assuming non-wizard non empty-random wish
   (as in makewish())
   
   These values were obtained by looking at mksobj()
   +3 Wishing for something in weapon_class
       +1 If that was something poisonable (darts, arrows)
   +4 Armor class
   +1 Wands
   +2 non-chargeable rings
   +1/+2 polymorph, teleportation, aggravete monster and hunger rings
   +1 spellbooks
   +1 potions and scrolls
   +2 amulets
   +1/+2 amulet of strangulation/change/restful/sleep
   +1 expensive camera/magic marker/tinning kit
   +2 can of grease
   +2 crystal ball
   +1 magic flute/magic harp/frost horn/fire horn/drum of earthquake
   +2 oil lamp
   +2/+3 candles 
   +1 kelp frond
   
   These are all in objnam.c, line numbers at the end
   +1 Wishing for a quantity and is a stackable item                    2473
   +1 Wishing for an enchantment and is a weapon/armor/ring             2494
   +1 Wishing for a wand of wishing (rn2(10) ? -1 : 0),                 2530
                          (can be tuned to get always (1:0) wands?
   +1 Wishing for an artifact (not if wishing your own quest artifact)  2965
   +1 Wishing for glass without specifying color                        2180
   +1 Wishing for gray stone                                          somewhere
   
   Some objects (like eggs) generate much more random numbers. Don't
   wish for them if you don't want your fountain to dry up.
*/

/* version  (day.month.year)
   v0.1     (28.2.2009)
   First released version.
*/

#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <cstdlib>
#include "srandom.hpp"

using namespace std;

int random_evaluations = 0;
RandomState rs;

inline char highc(char c)
{
    if (c >= 'a' && c <= 'z')
        return c - 'a' + 'A';
    return c;
};

inline char lowc(char c)
{
    if (c >= 'A' && c <= 'Z')
        return c - 'A' + 'a';
    return c;
};

void reset_random(unsigned int seed, unsigned int offset)
{
    rs.srandom(seed);
    unsigned int i1;
    for (i1 = 0; i1 < offset; i1++)
        rs.random();
    random_evaluations = offset;
}

int rn2(long x)
{
    random_evaluations++;
    return rs.random() % x;
}

int rne(int x, int experience_level)
{
	int tmp, utmp;

	utmp = (experience_level < 15) ? 5 : experience_level/3;
	tmp = 1;
	while (tmp < utmp && !rn2(x))
		tmp++;
		
	return tmp;
}

void nameTest(const string &artifact_name)
{
    int name_length = artifact_name.size();
    int n = rn2(name_length);
    char c1, c2;
    c1 = lowc(artifact_name[n]);
    do c2 = 'a' + rn2('z' - 'a'); while(c1 == c2);
}

bool unpredictabilityChecks(unsigned int seed, unsigned int random_state)
{
    // Don't allow rne()-calls in mkobj.c for weapons
    if (!rn2(11))
        return false;
    if (!rn2(10))
        return false;
        
    reset_random(seed, random_state);
    // Don't allow rne()-calls for armor
    if (!rn2(10))
        return false;
    if (!rn2(11))
        return false;
    if (!rn2(10))
        return false;
        
    reset_random(seed, random_state);
    if (!rn2(10))
        return false;
    if (!rn2(9))
        return false;

    return true;
}

bool demonGenerationTest(unsigned int dlevel)
{
    // Water demon
    int fate = rn2(30)+1;
    if (fate != 23)
        return false;
    
    // enexto_core() generates 1 random number (number of valid positions)
    rn2(8);
    // hp roll (for level 7 water demon, 7d8)
    rn2(8); rn2(8); rn2(8); rn2(8); rn2(8); rn2(8); rn2(8);
    // gender roll
    rn2(2);
    
    // We are too lazy to handle item generation rolls.
    // Just check that demon is not generated with items.
    
    int m_initweap_roll = rn2(12);
    if (m_initweap_roll < 6)
        return false;
        
    int offensive_item = rn2(75);
    if (7 > offensive_item)
        return false;
    
    int random_defensive_item = rn2(50);
    if (7 > random_defensive_item)
        return false;
        
    int random_misc_item = rn2(100);
    if (7 > random_misc_item)
        return false;
    
    // Wish
    unsigned int roll = rn2(100)+1;
    if (roll <= 80 + dlevel)
        return false;

    return true;
}

int namingWisher(unsigned int seed, 
                 unsigned int offset,
                 const string &naming,
                 int fountaindry,
                 int dlevel,
                 const string &output_file,
                 const string &fountain_file)
{
    // TODO: let the user decide max_offset
    unsigned int max_offset = 0xffffffff;
    if (max_offset < offset)
        max_offset = 0xffffffff;
    
    // Walking to a wall without moving generates exactly one random() call
    // (somewhere in hack.c, degrading engravings)
    
    bool bailout = false;
    unsigned int i1, i2;
    for (i1 = offset; i1 <= max_offset && !bailout; i1++)
    {
        if (i1 == 0xffffffff)
            bailout = true;
        
        reset_random(seed, offset);
        for (i2 = 0; i2 < (i1-offset); i2++)
            nameTest(naming);
		
		if (!demonGenerationTest(dlevel))
		    continue;
        
        int random_state = random_evaluations;
        
        // At this point, wish is guaranteed but we wan't to make sure that
        // user can easily predict the wish offset to prevent drying up the
        // fountain.
        
        if (fountaindry != -1 && !unpredictabilityChecks(seed, random_state))
            continue;
        
        // We get a wish. But do we preserve the fountain?
        int i2;
        vector<bool> fountain_preserves(10);
        for (i2 = 0; i2 < 10; i2++) // Show 10 wish offsets for now
        {
            reset_random(seed, random_state+i2);
            if (!rn2(3))
                fountain_preserves[i2] = false;
            else
                fountain_preserves[i2] = true;
        }
        
        if (fountaindry >= 2 && fountaindry <= 7 && 
            (!fountain_preserves[fountaindry] || !fountain_preserves[fountaindry-1] || !fountain_preserves[fountaindry+1] ||
             !fountain_preserves[fountaindry-2] || !fountain_preserves[fountaindry+2]))
            continue;
        
        cout << "Found wish after " << (i1-offset) << " namings." << endl;
        
        for (i2 = 0; i2 < 10; i2++)
            if (!fountain_preserves[i2])
                cout << "Fountain dries up with wish offset " << i2 << "." << endl;
            else
                cout << "Fountain does not dry up with wish offset " << i2 << "." << endl;
        
        // Write fountain preserve information to file, if user requested it
        if (fountain_file != "")
        {
            fstream fountain_stream(fountain_file.c_str(), fstream::out);
            if (fountain_stream.fail())
                cerr << "Failed to open " << fountain_file << endl;
            else
            {
                for (i2 = 0; i2 < 10; i2++)
                    fountain_stream << (int) fountain_preserves[i2] << endl;
                fountain_stream.close();
            }
        }
        
        // Write game commands to a file
        fstream output_stream(output_file.c_str(), fstream::out);
        if (output_stream.fail())
        {
            cerr << "Failed to open " << output_file << "." << endl;
            return -1;
        }
        
        for (i2 = 0; (unsigned int) i2 < i1-offset; i2++)
            output_stream << "#name\ny%" << naming << "\n   ";
        
        output_stream.close();
        return 0;
    }
    
    cout << "No match." << endl;
    return 0;
}

int wallWisher(unsigned int seed, 
               unsigned int offset,
               char wall_direction,
               int fountaindry,
               int dlevel,
               const string &output_file,
               const string &fountain_file)
{
    // TODO: let the user decide max_offset
    unsigned int max_offset = 0xffffffff;
    if (max_offset < offset)
        max_offset = 0xffffffff;
    
    // Walking to a wall without moving generates exactly one random() call
    // (somewhere in hack.c, degrading engravings)
    
    bool bailout = false;
    unsigned int i1;
    for (i1 = offset; i1 <= max_offset && !bailout; i1++)
    {
        if (i1 == 0xffffffff)
            bailout = true;
        
        reset_random(seed, i1);
		if (!demonGenerationTest(dlevel))
		    continue;
        
        int random_state = random_evaluations;
        
        // At this point, wish is guaranteed but we wan't to make sure that
        // user can easily predict the wish offset to prevent drying up the
        // fountain.
        
        if (fountaindry != -1 && !unpredictabilityChecks(seed, random_state))
            continue;
        
        // We get a wish. But do we preserve the fountain?
        int i2;
        vector<bool> fountain_preserves(10);
        for (i2 = 0; i2 < 10; i2++) // Show 10 wish offsets for now
        {
            reset_random(seed, random_state+i2);
            if (!rn2(3))
                fountain_preserves[i2] = false;
            else
                fountain_preserves[i2] = true;
        }
        
        if (fountaindry >= 2 && fountaindry <= 7 && 
            (!fountain_preserves[fountaindry] || !fountain_preserves[fountaindry-1] || !fountain_preserves[fountaindry+1] ||
             !fountain_preserves[fountaindry-2] || !fountain_preserves[fountaindry+2]))
            continue;
        
        cout << "Found wish after " << (i1-offset) << " moves to a wall." << endl;
        
        for (i2 = 0; i2 < 10; i2++)
            if (!fountain_preserves[i2])
                cout << "Fountain dries up with wish offset " << i2 << "." << endl;
            else
                cout << "Fountain does not dry up with wish offset " << i2 << "." << endl;
        
        // Write fountain preserve information to file, if user requested it
        if (fountain_file != "")
        {
            fstream fountain_stream(fountain_file.c_str(), fstream::out);
            if (fountain_stream.fail())
                cerr << "Failed to open " << fountain_file << endl;
            else
            {
                for (i2 = 0; i2 < 10; i2++)
                    fountain_stream << (int) fountain_preserves[i2] << endl;
                fountain_stream.close();
            }
        }
        
        // Write game commands to a file
        fstream output_stream(output_file.c_str(), fstream::out);
        if (output_stream.fail())
        {
            cerr << "Failed to open " << output_file << "." << endl;
            return -1;
        }
        
        for (i2 = 0; (unsigned int) i2 < i1-offset; i2++)
            output_stream << wall_direction;
        
        output_stream.close();
        return 0;
    }
    
    cout << "No match." << endl;
    return 0;
}
                


int main(int argc, char* argv[])
{
    unsigned int seed = 0, offset = 0;
    bool got_seed = false;
    unsigned char wall_move = 0;
    int dlevel = 1;
    string naming;
    int fountaindry = -1;
    
    string output_file("fountain_wish.output");
    string fountain_file;
    unsigned int u_argc = (unsigned int) argc;

    unsigned int i1;
    for (i1 = 1; i1 < u_argc; i1++)
    {
        string argument = argv[i1];
        if (argument == string("-s") || argument == string("--seed"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected seed after -s or --seed." << endl;
                return -1;
            }
            
            seed = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            got_seed = true;
            i1++;
        }
        else if (argument == string("--offset"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected offset after --offset." << endl;
                return -1;
            }
            
            offset = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--dlevel"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected dungeon level after --dlevel." << endl;
                return -1;
            }
            
            dlevel = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--name"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected artifact name after --name." << endl;
                return -1;
            }
            
            naming = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("--wall"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected direction after --wall." << endl;
                return -1;
            }
            
            if (string(argv[i1+1]).size() != 1)
            {
                cerr << "Direction can only have 1 character." << endl;
                return -1;
            }
            
            wall_move = argv[i1+1][0];
            i1++;
        }
        else if (argument == string("--fountaindry"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --fountaindry." << endl;
                return -1;
            }
            
            fountaindry = atoi(argv[i1+1]);
            i1++;
        }
        else if (argument == string("-o") || argument == string("--output"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -o or --output." << endl;
                return -1;
            }
            
            output_file = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("-f") || argument == string("--fountainfile"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -f or --fountainfile." << endl;
                return -1;
            }
            
            fountain_file = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("-h") || argument == string("--help"))
        {
            cout << "fountain_wish --help" << endl;
            cout << "fountain_wish -h" << endl;
            cout << "    See this help" << endl;
            cout << "fountain_wish --name engravename" << endl;
            cout << "    If you want to use #naming method to generate more numbers," << endl;
            cout << "    specify the artifact name you use." << endl;
            cout << "fountain_wish --wall direction" << endl;
            cout << "    If you want to generate random numbers by walking to a wall," << endl;
            cout << "    specify the direction you want to move." << endl;
            cout << "    Direction can be any character. It is used when output file is" << endl;
            cout << "    being created." << endl;
            cout << "fountain_wish --difficulty dungeonlevel" << endl;
            cout << "    Specify your current level difficulty. Defaults to 1." << endl;
            cout << "    Level difficulty is the average of dungeon level and your experience level." << endl;
            cout << "fountain_wish --seed value" << endl;
            cout << "fountain_wish -s value" << endl;
            cout << "    Use the given seed. Must be used." << endl;
            cout << "fountain_wish --fountaindry wishoffset" << endl;
            cout << "    Ensures that a wish offset will not dry up the fountain." << endl;
            cout << "fountain_wish --offset value" << endl;
            cout << "    Use the given offset for random() -calls." << endl;
            cout << "    This is the value you get from seed_by_naming, the end offset." << endl;
            cout << "    Defaults to 0." << endl;
            cout << "fountain_wish -o filename" << endl;
            cout << "fountain_wish --output filename" << endl;
            cout << "    To help you with naming, a file is created with contents you can send" << endl;
            cout << "    to your game before quaffing." << endl;
            cout << "    When you use naming method, % is used for the item letter. You need to" << endl;
            cout << "    replace that with the actual item letter." << endl;
            cout << "    Defaults to fountain_wish.output." << endl;
            cout << "fountain_wish -f filename" << endl;
            cout << "fountain_wish --fountainfile filename" << endl;
            cout << "    Writes fountain preserve information in a file." << endl;
            cout << "    Each line represents if fountain dries up at that wish offset." << endl;
            cout << "    Mostly useful for automated tools that need the information in easily parseable form." << endl;
            cout << "Note:" << endl;
            cout << "    If you have the AoY, you cannot get wishes from a fountain." << endl;
            cout << "Usage example:" << endl;
            cout << "fountain_wish -s 1239138221 --offset 89 --output paste_me.txt" << endl;
            cout << endl;
            return 0;
        }
        else
        {
            cerr << "Unknown argument \"" << argv[i1] << "\"." << endl;
            cerr << "Use --help or -h to see accepted arguments." << endl;
            return -1;
        }
    };
    
    if (!got_seed)
    {
        cerr << "You must specify seed." << endl;
        return -1;
    }
    
    // Cannot use wall moving and naming at the same time
    if (wall_move && naming != "")
    {
        cerr << "Cannot use --wall with --name." << endl;
        return -1;
    }
    if (!wall_move && naming == "")
    {
        cerr << "You must use either --wall or --name." << endl;
        return -1;
    }
    
    int result = 0;
    if (wall_move)
        result = wallWisher(seed, offset, wall_move, fountaindry, dlevel, output_file, fountain_file);
    else
        result = namingWisher(seed, offset, naming, fountaindry, dlevel, output_file, fountain_file);
    
    return result;
};


