/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <cstdlib>
#include "srandom.hpp"
#include "caches.hpp"

using namespace std;

inline char highc(char c)
{
    if (c >= 'a' && c <= 'z')
        return c - 'a' + 'A';
    return c;
};

inline char lowc(char c)
{
    if (c >= 'A' && c <= 'Z')
        return c - 'A' + 'a';
    return c;
};

unsigned int random_evaluations = 0;

string name_test(const vector<string> &names, RandomState &rs)
{
    int name_length = names[0].size();
    string result = names[0];
    int n = rs.random() % name_length;
    ++random_evaluations;
    
    char c1, c2;
    c1 = lowc(names[0][n]);
    do 
    {
        c2 = 'a' + (rs.random() % ('z' - 'a'));
        ++random_evaluations;
    } while(c1 == c2);
    result[n] = (names[0][n] == c1) ? c2 : highc(c2);
    
    return result;
}

void name_test2(const vector<string> &names, string &result, RandomState &rs)
{
    int name_length = names[0].size();
    result = names[0];
    int n = rs.random() % name_length;
    ++random_evaluations;
    
    char c1, c2;
    c1 = lowc(names[0][n]);
    do 
    {
        c2 = 'a' + (rs.random() % ('z' - 'a')); 
        ++random_evaluations;
    } while(c1 == c2);
    result[n] = (names[0][n] == c1) ? c2 : highc(c2);
}

bool name(const vector<string> &names, 
          const int &index,
          RandomState &rs)
{
    int name_length = names[0].size();
    int n = rs.random() % name_length;
    ++random_evaluations;
    
    char c1, c2;
    c1 = lowc(names[0][n]);
    do 
    {
        c2 = 'a' + (rs.random() % ('z' - 'a')); 
        ++random_evaluations;
    } while(c1 == c2);
    char result = (names[0][n] == c1) ? c2 : highc(c2);
        
    if (result != names[index+1][n])
        return false;
    
    return true;
}

/* The maximum number of files open on linux systems is usually 1024.
   The number returned by this function will be the number of cache files
   open. The cache builder can work around the limit but that's slow. */
unsigned int getMaxHashNumbers()
{
    return 900;
}

unsigned int getHashNumber(const string &n1, const string &n2)
{
    unsigned int result;
    Cache::hashString(n1 + n2, result);
    return result % (900);
}

int buildCache(const string &cache_name, const vector<string> &names, unsigned int start_index, unsigned int end_index)
{
    string original_name = names[0];
    string result, result2;
    
    bool bailout = false;
    Cache cache;
    if (!cache.newCache(cache_name, getMaxHashNumbers()))
    {
        cerr << "Failed to create a new cache." << endl;
        return -1;
    }
    
    time_t start_time0 = time(0);
    time_t last_time0 = start_time0;
    
    RandomState rs;
    unsigned int index;
    for (index = start_index; index <= end_index && !bailout; index++)
    {
        if (index == 0xffffffff)
            bailout = true;
        
        rs.srandom(index);
        name_test2(names, result, rs);
        name_test2(names, result2, rs);
        unsigned int hash_number = getHashNumber(result, result2);
        cache.addSeed(hash_number, index);
        
        if (!(index % 1000))
            if (last_time0 != time(0))
            {
                last_time0 = time(0);
                
                unsigned long eta = last_time0 - start_time0;
                unsigned long finished_for_now = (unsigned long )index - (unsigned long) start_index;
                eta = (unsigned int) (((double) ((unsigned long) end_index - (unsigned long) index + 1) / (double) (finished_for_now)) * (double) eta);
                cout << "At seed " << index << ", estimated time " << eta << " seconds." << endl;
            };
    }
    
    cache.closeCache();
    return 0;
}

bool SingleCandidate = false;

void crushWithoutCache(fstream &output_stream,
                       unsigned int start_index, 
                       unsigned int end_index, 
                       unsigned int start_offset,
                       unsigned int end_offset,
                       const vector<string> &names)
{    
    unsigned int i1, i2, i3, index, names_len = names.size();
    bool bailout = false;
    
    time_t start_time0 = time(0);
    time_t last_time0 = start_time0;
    
    RandomState return_rs, rs;
    for (index = start_index; index <= end_index && !bailout; index++)
    {
        if (index == 0xffffffff)
            bailout = true;
        
        return_rs.srandom(index);
        for (i2 = start_offset; i2 <= end_offset; i2++)
        {
            rs = return_rs;
            for (i3 = 0; i3 < i2; i3++)
                rs.random();
            random_evaluations = i2;
                
            for (i1 = 0; i1 < names_len-1; i1++)
                if (!name(names, i1, rs))
                    break;
                    
            if (i1 >= names_len-1)
            {
                cout << "New candidate: " << index << ", offset " << i2 << ", end offset " << random_evaluations << endl;
                output_stream << index << " " << i2 << " " << random_evaluations << endl;
                output_stream.flush();            
                
                if (SingleCandidate)
                    return;
            }
        }
        
        if (!(index % 1000))
            if (last_time0 != time(0))
            {
                last_time0 = time(0);
                
                unsigned long eta = last_time0 - start_time0;
                unsigned long finished_for_now = (unsigned long )index - (unsigned long) start_index;
                eta = (unsigned int) (((double) ((unsigned long) end_index - (unsigned long) index + 1) / (double) (finished_for_now)) * (double) eta);
                cout << "At seed " << index << ", estimated time " << eta << " seconds." << endl;
            };
    };
}

void crush(fstream &output_stream,
           const string &use_cache,
           unsigned int start_index, 
           unsigned int end_index, 
           unsigned int start_offset,
           unsigned int end_offset,
           const vector<string> &names)
{
    if (use_cache == "")
    {
        crushWithoutCache(output_stream, start_index, end_index, start_offset, end_offset, names);
        return;
    }
        
    /* crush with cache, open the cache */
    Cache cache;
    if (!cache.openCache(use_cache, getMaxHashNumbers()))
    {
        cerr << "Could not open cache." << endl;
        return;
    };
    
    unsigned int i1, i2, i3, names_len = names.size();
    
    time_t start_time0 = time(0);
    time_t last_time0 = start_time0;
    
    unsigned int hash_number = getHashNumber(names[1], names[2]);
    
    vector<unsigned int> seeds;
    cache.readCacheFile(hash_number, seeds);

    RandomState return_rs, rs;    
    vector<unsigned int>::iterator index_i;
    for (index_i = seeds.begin(); index_i <= seeds.end(); index_i++)
    {
        unsigned int index = *index_i;
        
        return_rs.srandom(index);

        for (i2 = start_offset; i2 <= end_offset; i2++)
        {
            rs = return_rs;
            for (i3 = 0; i3 < i2; i3++)
                rs.random();
            random_evaluations = i2;
            
            for (i1 = 0; i1 < names_len-1; i1++)
                if (!name(names, i1, rs))
                    break;
                    
            if (i1 >= names_len-1)
            {
                cout << "New candidate: " << index << ", offset " << i2 << ", end offset " << random_evaluations << endl;
                output_stream << index << " " << i2 << " " << random_evaluations << endl;
                output_stream.flush();            
                
                if (SingleCandidate)
                    return;
            }
        }
        
        if (!(index % 1000))
            if (last_time0 != time(0))
            {
                last_time0 = time(0);
                
                unsigned long eta = last_time0 - start_time0;
                unsigned long finished_for_now = (unsigned long )index - (unsigned long) start_index;
                eta = (unsigned int) (((double) ((unsigned long) end_index - (unsigned long) index + 1) / (double) (finished_for_now)) * (double) eta);
                cout << "At seed " << index << ", estimated time " << eta << " seconds." << endl;
            };
    };
}

void showNamings(unsigned int seed, unsigned int offset, int num_names, const vector<string> &names)
{
    RandomState rs;
    rs.srandom(seed);
    int i1;
    for (i1 = 0; i1 < (int) offset; i1++)
        rs.random();    

    for (i1 = 0; i1 < num_names; i1++)
    {
        string r = name_test(names, rs);
        cout << r << endl;
    }
};

int readConfigurationFile(const string &conf_filename, vector<string> &names)
{
    fstream conf_stream(conf_filename.c_str(), fstream::in);
    if (conf_stream.fail())
    {
        cerr << "Could not open " << conf_filename << "." << endl;
        return -1;
    };

    // Configuration file format is:
    // "correct artifact name" (see include/artilist.h in nethack source)
    // "first naming result"
    // "second naming result"
    // "third naming result"
    // etc.
    
    // e.g.
    // "The Staff of Aesculapius"
    // "The Staff of Alsculapius"
    // "The Staff of Aedculapius"
    // "The Staff of Aesculapins"
    
    while(!conf_stream.eof())
    {
        string dummy, result;
        // left quote
        getline(conf_stream, dummy, '\"');
        if (conf_stream.eof())
            break;
        // naming result
        getline(conf_stream, result, '\"');
        // newline
        getline(conf_stream, dummy, '\n');
            
        names.push_back(result);
    };
    
    if (names.size() < 2)
    {
        cout << "Every seed is a candidate." << endl;
        return 0;
    }
    
    // All names must be of the same length
    // Also check that all names have exactly one mistaken letter
    unsigned int len = names[0].size();
    vector<string>::iterator i1;
    for (i1 = names.begin(); i1 != names.end(); i1++)
    {
        if (i1->size() != len)
        {
            cerr << "All names should have the same length. Check your configuration." << endl;
            return -1;
        }
        
        // First name is correct
        if (i1 == names.begin())
            continue;
        
        unsigned int number_of_mistakes = 0, i2;
        for (i2 = 0; i2 < i1->size(); i2++)
            if ((*i1)[i2] != names[0][i2])
                number_of_mistakes++;
        
        if (number_of_mistakes != 1)
        {
            cerr << "Naming result \"" << (*i1) << "\" has " << number_of_mistakes << " wrong letters. It should be 1." << endl;
            return -1;
        }
    }
    
    return 0;
};

int main(int argc, char* argv[])
{
    unsigned int start_index = 0, end_index = 1000000;
    unsigned int start_offset = 0, end_offset = 0;
    
    string conf_filename("seed_by_naming.conf");
    string output_filename("seed_by_naming.output");
    string use_cache, build_cache;
    
    vector<string> names;
    bool loaded_configuration = false;
    
    unsigned int u_argc = (unsigned int) argc;

    unsigned int i1;
    for (i1 = 1; i1 < u_argc; i1++)
    {
        string argument = argv[i1];        
        if (argument == string("-i") || argument == string("--inspect"))
        {
            if (i1 >= u_argc-2)
            {
                cerr << "Expected seed and offset after -i or --inspect." << endl;
                return -1;
            }
            if (!loaded_configuration)
            {
                int result = readConfigurationFile(conf_filename, names);
                if (result)
                    return -1;
                loaded_configuration = true;
            }
            
            unsigned int seed = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            unsigned int offset = (unsigned int) strtoul(argv[i1+2], NULL, 10);
            showNamings(seed, offset, 30, names);
            return 0;
        }
        else if (argument == string("--buildcache"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected cache name after --buildcache." << endl;
                return -1;
            }
            
            build_cache = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("--usecache"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected cache name after --usecache." << endl;
                return -1;
            }
            
            use_cache = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("-c") || argument == string("--conf"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -c or --conf." << endl;
                return -1;
            };
            conf_filename = string(argv[i1+1]);
            i1++;
            if (!loaded_configuration)
            {
                int result = readConfigurationFile(conf_filename, names);
                if (result)
                    return -1;
                loaded_configuration = true;
            }
            else
            {
                cerr << "Cannot have more than one configuration file." << endl;
                return -1;
            }
        }
        else if (argument == string("-o") || argument == string("--output"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -o or --output." << endl;
                return -1;
            };
            output_filename = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("--offsetstart"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --offsetstart." << endl;
                return -1;
            }
            
            start_offset = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--offsetend"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --offsetend." << endl;
                return -1;
            }
            
            end_offset = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--indexstart"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --indexstart." << endl;
                return -1;
            };
            start_index = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--indexend"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --indexend." << endl;
                return -1;
            };
            end_index = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--singlecandidate"))
            SingleCandidate = true;
        else if (argument == string("-h") || argument == string("--help"))
        {
            cout << "seed_by_naming --conf filename" << endl;
            cout << "seed_by_naming -c filename" << endl;
            cout << "    Use your own configuration file. Defaults to seed_by_naming.conf." << endl;
            cout << "seed_by_naming --singlecandidate" << endl;
            cout << "    Stop after finding a candidate." << endl;
            cout << "seed_by_naming --buildcache cachename" << endl;
            cout << "    Rather than finding the seed, build caches for the first engraving in the configuration." << endl;
            cout << "    The cache takes up to 16 gigabytes for 2^32 seeds. Building the cache usually takes hours." << endl;
            cout << "seed_by_naming --usecache cachename" << endl;
            cout << "    Use a cache for finding the values." << endl;
            cout << "seed_by_naming --output filename" << endl;
            cout << "seed_by_naming -o filename" << endl;
            cout << "    Output candidates in a file. Defaults to seed_by_naming.output." << endl;
            cout << "seed_by_naming --indexstart value" << endl;
            cout << "    Start seed forcing from this index (default 0)." << endl;
            cout << "seed_by_naming --indexend value" << endl;
            cout << "    Start seed forcing from this index (default 0)." << endl;
            cout << "seed_by_naming --help" << endl;
            cout << "seed_by_naming -h" << endl;
            cout << "    See this help" << endl;
            cout << "seed_by_naming --offsetstart value" << endl;
            cout << "seed_by_naming --offsetend value" << endl;
            cout << "    Set the number of random() -calls checked per seed." << endl;
            cout << "    Default is equivalent to --offsetstart 0 --offsetend 0." << endl;
            cout << "seed_by_naming --inspect seed" << endl;
            cout << "seed_by_naming -i seed" << endl;
            cout << "    Shows 30 naming results at the given seed." << endl;
            cout << "Usage example:" << endl;
            cout << "seed_by_naming -c namings.conf -o namings.candidates --indexend 1000000000" << endl;
            cout << endl;
            return 0;
        }
        else
        {
            cerr << "Unknown argument \"" << argv[i1] << "\"." << endl;
            cerr << "Use --help or -h to see accepted arguments." << endl;
            return -1;
        }
    };

    if (!loaded_configuration)
    {
        int result = readConfigurationFile(conf_filename, names);
        if (result)
            return -1;
    }
    
    if (build_cache != "")
        return buildCache(build_cache, names, start_index, end_index);
    
    fstream output_stream(output_filename.c_str(), fstream::out);
    if (output_stream.fail())
    {
        cerr << "Failed to open " << output_filename << " for writing." << endl;
        return -1;
    };
    
    crush(output_stream, use_cache, start_index, end_index, start_offset, end_offset, names);
    
    output_stream.close();
    return 0;
}

