/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifndef srandom_h
#define srandom_h

#include <vector>

/* Class, that implements srandom()/srand()/random()/rand() as they are
   in the GNU C library. Easy to copy around C++ style. */
   
/* When compiled with optimizations, this class is faster than 
   srandom()/random() */

/* Define this, if you want your 64-bit CPU to emulate
   64-bit srandom()/random(). There's a bug in GNU C library, that
   causes different random values to be returned with half of the seeds.
   If you undefine this, it calculates values like any 32-bit GNU C library. */
//#define EMULATE_64BIT_BUG

class RandomState
{
    private:
        uint32_t* r;
        uint32_t* r_2;
        uint32_t* r_30;
        uint32_t* r_33;
        uint32_t* r_end;
        
        void r_forward()
        {
            ++r_2;
            ++r_30;
            ++r_33;
            if (r_2 == r_end)
                r_2 = r;
            if (r_30 == r_end)
                r_30 = r;
            if (r_33 == r_end)
                r_33 = r;
            
            (*r_33) = (*r_30) + (*r_2);
        }
        
        void setupPointers()
        {
            r_2 = &r[2];
            r_30 = &r[30];
            r_33 = &r[33];
            r_end = &r[34];
        }
        
    public:
        RandomState()
        {
            r = new uint32_t[34];
            setupPointers();
        };
        
        RandomState(unsigned int s)
        {
            r = new uint32_t[34];
            setupPointers();
            srandom(s);
        };
        ~RandomState()
        {
            delete[] r;
        }
        RandomState(const RandomState &rs)
        {
            r = new uint32_t[34];
            setupPointers();
            int i1;
            for (i1 = 0; i1 < 34; i1++)
                r[i1] = rs.r[i1];
            r_2 = (rs.r_2 - rs.r) + r;
            r_30 = (rs.r_30 - rs.r) + r;
            r_33 = (rs.r_33 - rs.r) + r;
        }
        RandomState& operator=(const RandomState &rs)
        {
            if (this == &rs)
                return (*this);
            int i1;
            for (i1 = 0; i1 < 34; i1++)
                r[i1] = rs.r[i1];
            r_2 = (rs.r_2 - rs.r) + r;
            r_30 = (rs.r_30 - rs.r) + r;
            r_33 = (rs.r_33 - rs.r) + r;
            return (*this);
        }
        
        void srandom(unsigned int s)
        {
            if (s == 0) s = 1;
            setupPointers();
            
            r[0] = s;
            int i1;
            
            #ifdef EMULATE_64BIT_BUG
            long int w = s;
            for (i1 = 1; i1 <= 30; i1++)
            {
                long int hi = w / 127773;
                long int lo = w % 127773;
                w = 16807 * lo - 2836 * hi;
                if (w < 0)
                    w += 2147483647;
                r[i1] = w;
            }
            #else
            int32_t w = s;
            for (i1 = 1; i1 <= 30; i1++)
            {
                int32_t hi = w / 127773;
                int32_t lo = w % 127773;
                w = 16807 * lo - 2836 * hi;
                if (w < 0)
                    w += 2147483647;
                r[i1] = w;
            }
            #endif
            r[31] = r[0];
            r[32] = r[1];
            r[33] = r[2];
            
            for (i1 = 0; i1 < 311; i1++)
                r_forward();
        };
        
        long int random()
        {
            unsigned int result = (*r_33) >> 1;
            r_forward();
            return (long int) result;
        }
};

#endif

